# BTC.py - automatyzacja zapytań deanonimizacyjnych w sieci Bitcoin

# koncepcja     : Przemysław RODWALD [p.rodwald@amw.gdynia.pl]
# implementacja : Nicola KOŁAKOWSKA  [n.kolakowska@outlook.com]

import requests
import json
import hashlib
import pandas as pd
import time
import random
import string
from difflib import SequenceMatcher

#Instalacja bibliotek:
#pip install requests
#pip install pandas

main_address, direction, sites, iterate, max_noa = input("Enter address, direction [IN/OUT], sites [SWGIC], iterate [1-], max number of addressess [1-], \
                                                         e.g. : 1AXZ1SwA2uW2cYoKmH23XgDyWGdEf3RwRB OUT SWGIC 2 50: \n").split()

iterate = int(iterate)
max_noa = int(max_noa)

print("Please be patient. Query in progress...")

#KLUCZE
#(200 r/1 h)
apikey_blockcypher = "" #miejsce na klucz api pozyskany z blockcypher
#(no limit)
apikey_sydeus = "" #miejsce na klucz api pozyskany z sydeus
#(5000 r/5 min)
apikey_chainalysis = { #nagłówek dla chainalysis wraz z kluczem api
    'X-API-Key':'', #miejsce na klucz api pozyskany z chainalysis
    'Accept': 'application/json'}
#(1000 r/1 h)
apikey_graphsense = { #nagłówek dla graphsense wraz z kluczem autoryzacji    
    'accept': 'application/json',
    'Authorization': ''} #miejsce na klucz api pozyskany z graphsense
#(1000 r/1 day)
apikey_ikna =  { #nagłówek dla ikna.io z kluczem api
     'accept': 'application/json',
     'Authorization': ''} #miejsce na klucz api pozyskany z ikna
#walletexplorer (2 r/1 s)

address_list = [] #lista adresów, która zbiera adresy z zahashowanych transakcji
address_results = [[]] #lista adresów dla danej epoki
finish_results = [] #lista tablic typu dataframe dla wszystkich epok
all_of_addresses = ['ALL'] #lista wszystkich dotychczasowych adresów
temp = [] #zastępcza lista

      
def similar(a, b): #funkcja do sprawdzania podobieństwa dwóch stringów
    return SequenceMatcher(None, a, b).ratio()

def countdown(t): #funkcja do odliczania podanego czasu w przypadku przekroczenia limitów
    print("Time left:")
    while t:
        hours, mins = divmod(t, 60) #układ czasu godzina, minuty
        timer = '{:02d}:{:02d}\n'.format(hours, mins) #format czasomierza
        print(timer, end="\r")
        time.sleep(600) #uspienie programu na 10 minut
        t -= 10 #odjęcie wartosci
        
def limits():
    print("Limits have been exceeded!")
    print("Current epoch: ", a)
    countdown(60) #odliczanie 60 minut
    print("Resumed!")
    
def unique(list1, list2): #funkcja do utworzenia listy z unikatowymi wartosciami
    for x in list1: 
        if x not in list2:
            list2.append(x)
            
def hash_list_blockchain(add): #funkcja do utworzenia listy zahashowanych transakcji
    global apikey_blockcypher
    try:
        h1 = requests.get("https://blockchain.info/rawaddr/" + add).text
        if "<!DOCTYPE html>" in h1:
            print("Blockchain limits reached!")
            limits()
            if apikey_blockcypher == "":
                return hash_list_blockchain(add)
            else:
                return hash_list_blockcypher(add)
        else:
            h1 = json.loads(h1)
            txs = h1.get('txs')
            check = len(txs)
            if check < 100:
                txs = [x['hash'] for x in txs if 'hash' in x]
                txs2 = [] #tablica dla unikatowych zahashowanych transakcji
                unique(txs, txs2) #utworzenie listy unikatowych zahashowanych transakcji
                return txs2 #pobranie listy hashy
            else:
                txs = [x['hash'] for x in txs if 'hash' in x]
                txs2 = []
                txs2.extend(txs)
                offset = "100"
                txs3 = []
                while(check >= 100):
                    time.sleep(10)
                    h1 = requests.get("https://blockchain.info/rawaddr/" + add + "?offset=" + offset).text
                    h1 = json.loads(h1)
                    txs = h1.get('txs')
                    txs = [x['hash'] for x in txs if 'hash' in x]
                    check = len(txs)
                    txs2.extend(txs)
                    offset = int(offset)
                    offset += 100
                    offset = str(offset)
                unique(txs2, txs3) #utworzenie listy unikatowych zahashowanych transakcji
                return txs3 
    except requests.exceptions.RequestException as e:
        if apikey_blockcypher == "":
            print("Blockchain doesn't work!")
            raise SystemExit(e)
        else:
            limits()
            return hash_list_blockcypher(add)
        
def hash_list_blockcypher(add):
    global apikey_blockcypher
    try:
        h1 = requests.get("https://api.blockcypher.com/v1/btc/main/addrs/" + add +"?token=" + apikey_blockcypher).json() #detale dotyczące danego adresu
        if 'error' in h1.keys():
            print("Blockcypher limits reached!")
            return hash_list_blockchain(add)             
        else:    
            txrefs = h1.get('txrefs') #informacje na temat zahashowanych transakcji
            if len(txrefs) >= 100:
                return hash_list_blockchain(add)
            else:
                txrefs = [x['tx_hash'] for x in txrefs if 'tx_hash' in x] #pobranie i utworzenie listy zahashowanych transakcji 
                txrefs2 = []
                unique(txrefs, txrefs2)
                return txrefs2
    except requests.exceptions.RequestException:
        print("Blockcypher doesn't work!")
        apikey_blockcypher = ""
        return hash_list_blockchain(add)

def blockchain_addresses(hashes, address, tx_output_b):
    try:
        trans_details = requests.get("https://blockchain.info/rawtx/" + hashes).text
        if "<!DOCTYPE html>" in trans_details:
            print("Blockchain limits reached!")
            limits()
            if apikey_blockcypher == "":
                return blockchain_addresses(hashes, address, tx_output_b)
            else:
                return blockcypher_addresses(hashes, address, tx_output_b)
        else:
            trans_details = json.loads(trans_details)
            if "IN" in direction:
                input_b = trans_details.get('inputs')
                input_b = [x['prev_out'] for x in input_b if 'prev_out' in x]
                input_b = [x['addr'] for x in input_b if 'addr' in x]
                if address not in input_b:
                    tx_output_b.extend(input_b)
            else: #OUT 
                output_b = trans_details.get('out')
                output_b = [x['addr'] for x in output_b if 'addr' in x]
                if address not in output_b:
                    tx_output_b.extend(output_b)
    except requests.exceptions.RequestException as e:
        if apikey_blockcypher == "":
            print("Blockchain doesn't work!")
            raise SystemExit(e)
        else:
            limits()
            return blockcypher_addresses(hashes, address, tx_output_b)

def blockcypher_addresses(hashes, address, tx_output_b):
    global apikey_blockcypher
    try:
        trans_details = requests.get("https://api.blockcypher.com/v1/btc/main/txs/" + hashes + "?token=" + apikey_blockcypher).json() #informacje na temat konkretnej zahashowanej transakcji
        if 'error' in trans_details.keys():
            print("Blockcypher limits reached!")
            return blockchain_addresses(hashes, address, tx_output_b)             
        else:    
            if "IN" in direction:
                input_b = trans_details.get('inputs')
                input_b = [x['addresses'] for x in input_b if 'addresses' in x]
                input_b = [item for sublist in input_b for item in sublist]
                if address not in input_b:
                    tx_output_b.extend(input_b)               
            else:
                output_b = trans_details.get('outputs') 
                output_b = [x['addresses'] for x in output_b if 'addresses' in x]
                output_b = [item for sublist in output_b for item in sublist] 
                if address not in output_b:
                    tx_output_b.extend(output_b)
    except requests.exceptions.RequestException:
        print("Blockcypher doesn't work!")
        apikey_blockcypher = ""
        return blockchain_addresses(hashes, address, tx_output_b) 


def str_to_hash(add):
    str2hash = "nicola" + add + apikey_sydeus #hasło, główny adres i klucz dostępu do api do strony sydeus
    hash_result = hashlib.md5(str2hash.encode()) #ciąg znaków do zakodowania
    hash_result = hash_result.hexdigest() #format szesnastkowy
    return hash_result

def sydeus_add(add, list1, hash_result):
    try:
        sydeus_add = requests.get("https://sydeus.rodwald.pl/api.php?address=" + add + "&apikey=" + apikey_sydeus + "&crc=" + hash_result).json() #wysłanie zapytania do strony sydeus o informacje na temat konkretnego adresu
        sydeus_add = sydeus_add.get('sydeus')
        if 'success' in sydeus_add.values(): #jeżeli znaleziono dane
            list1.append(sydeus_add.get('entity'))
        elif 'failure' in sydeus_add.values():
            list1.append("")
        elif 'incorrect apikey' in sydeus_add.values(): #jeżeli nie znaleziono danych
            list1.append("incorrect")
        elif 'incorrect BTC address format' in sydeus_add.values():
            print("Incorrect address!")
            raise SystemExit()
    except requests.exceptions.RequestException:
        list1.append("???")

def walletexplorer_add(add, list1):
    try:
        walletexplorer_add = requests.get("http://www.walletexplorer.com/api/1/address-lookup?address=" + add + "&caller=" + "".join(random.choices(string.ascii_uppercase + string.digits, k=20))).json() #wysłanie zapytania do strony walletexplorer o informacje na temat konkretnego adresu
        if 'label' in walletexplorer_add.keys(): #jeżeli w kluczach znajduje się słowo label
            list1.append(walletexplorer_add.get('label')) #dodanie do listy adresu z nazwą
        elif False in walletexplorer_add.values():
            print("Incorrect address!")
            raise SystemExit()
        else:
            list1.append("")
    except requests.exceptions.RequestException:
        list1.append("???")

def graphsense_add(add, list1):
    try:
        graphsense_add = requests.get('https://api.graphsense.info/btc/addresses/' + add + '?include_tags=true', headers = apikey_graphsense).json()
        if "API rate limit exceeded" in graphsense_add.values():
            list1.append("???")
        elif "Not Found" in graphsense_add.values():
            print("Incorrect address!")
            raise SystemExit()
        else:
            if 'Unauthorized' in graphsense_add.values():
                list1.append("incorrect")
            else:
                graph = graphsense_add.get('entity')
                graph = str(graph)
                graphsense_add = requests.get('https://api.graphsense.info/btc/entities/' + graph + '/tags?level=entity&pagesize=20', headers = apikey_graphsense).json()
                graphsense_add = graphsense_add.get('address_tags')
                if not graphsense_add: #jeżeli brak danych to przejscie dalej
                    list1.append("")
                    pass
                else:
                    for i in range(len(graphsense_add)):
                        graph2 = graphsense_add[i]
                        if graph2.get('label') == "":
                            if i == (len(graphsense_add) - 1):
                                list1.append("")
                                break
                            continue
                        else:
                            list1.append(graph2.get('label'))
                            break
    except requests.exceptions.RequestException:
        list1.append("???")

def ikna_add(add, list1):
    try:
        ikna_add = requests.get('https://api.ikna.io/btc/addresses/' + add + '?include_tags=true', headers = apikey_ikna).json()
        if "API rate limit exceeded" in ikna_add.values():
            list1.append("???")
        elif "Not Found" in ikna_add.values():
            print("Incorrect address!")
            raise SystemExit()
        else:
            if 'Unauthorized' in ikna_add.values():
                list1.append("incorrect")
            else:
                ikna = ikna_add.get('entity')
                ikna = str(ikna)
                ikna_add = requests.get('https://api.ikna.io/btc/entities/' + ikna + '/tags?level=entity&pagesize=20', headers = apikey_ikna).json()
                ikna_add = ikna_add.get('address_tags')
                if not ikna_add: #jeżeli brak danych to przejscie dalej
                    list1.append("")
                    pass
                else:
                    for i in range(len(ikna_add)):
                        ikna2 = ikna_add[i]
                        if ikna2.get('label') == "":
                            if i == (len(ikna_add) - 1):
                                list1.append("")
                                break
                            continue
                        else:
                            list1.append(ikna2.get('label'))
                            break
                        
    except requests.exceptions.RequestException:
        list1.append("???")
        
def chainalysis_add(add, list1):
    try:
        chainalysis_add = requests.get('https://public.chainalysis.com/api/v1/address/' + add, headers = apikey_chainalysis) #wysłanie zapytania do strony chainalysis o informacje na temat konkretnego adresu
        chainalysis_add = json.loads(chainalysis_add.text) #informacje w formie json
        if 'Invalid API Key' in chainalysis_add.values():
            list1.append("incorrect")
        else:
            chainalysis_add = chainalysis_add.get('identifications')
            if len(chainalysis_add) > 0:
                chainalysis_add = dict(chainalysis_add[0])
                chainalysis_add = chainalysis_add.get('category')
                list_chain = ["S"]
                list1.append(list_chain) #dodanie informacji do listy
            else:
                list1.append("_")
    except requests.exceptions.RequestException:
        list1.append("???")
            

def crypto_function(address): #główna funkcja, w której wprowadzane są adresy i tworzą się nowe tablice z adresami
    global all_of_addresses, address_results, finish_results, list_of_main_addresses, a, n, sites, main_address, \
            temp, apikey_blockcypher, apikey_sydeus, apikey_graphsense, apikey_chainalysis, apikey_ikna
    
    data_sydeus = [] #lista do przechowywania informacji o adresie ze strony sydeus
    data_walletexplorer = [] #lista do przechowywania informacji o adresie ze strony walletexplorer
    data_graphsense = [] #lista do przechowywania informacji o adresie ze strony graphsense
    data_ikna = [] #lista do przechowywania informacji o adresie ze strony ikna
    data_chainalysis = [] #lista do przechowywania informacji o adresie ze strony chainalysis
    

    df = pd.DataFrame(columns=['Address'])
    df.loc[0, 'Address'] = address
    #sydeus
    if "S" in sites:
        #hash
        hash_result = str_to_hash(address)        
        sydeus_add(address, data_sydeus, hash_result)
        if 'incorrect' not in data_sydeus:
            if 'Sydeus' in df.columns:
                df.loc[0, 'Sydeus'] = data_sydeus
            else:
                df['Sydeus'] = data_sydeus
        else:
            print("Invalid or missing apikey for Sydeus!")
            sites = sites.translate({ord('S'): None})
         
    #walletexplorer
    if "W" in sites or sites == '':
        walletexplorer_add(address, data_walletexplorer)
        if 'Walletexplorer' in df.columns:
            df.loc[0, 'Walletexplorer'] = data_walletexplorer
        else:
            df['Walletexplorer'] = data_walletexplorer
         
    #graphsense
    if "G" in sites: 
        graphsense_add(address, data_graphsense)
        if 'incorrect' not in data_graphsense:
            if 'Graphsense' in df.columns:
                df.loc[0, 'Graphsense'] = data_graphsense
            else:
                df['Graphsense'] = data_graphsense
        else:
            print("Invalid or missing apikey for Graphsense!")
            sites = sites.translate({ord('G'): None})
     
    #ikna
    if "I" in sites:
        ikna_add(address, data_ikna)
        if 'incorrect' not in data_ikna:
            if 'Ikna' in df.columns:
                df.loc[0, 'Ikna'] = data_ikna
            else:
                df['Ikna'] = data_ikna
        else:
            print("Invalid or missing apikey for Ikna!")
            sites = sites.translate({ord('I'): None})
         
    #chainalysis
    if "C" in sites:
        chainalysis_add(address, data_chainalysis)
        if 'incorrect' not in data_chainalysis:
            if 'Chainalysis' in df.columns:
                df.loc[0, 'Chainalysis'] = data_chainalysis
            else:
                df['Chainalysis'] = data_chainalysis
        else:
            print("Invalid or missing apikey for Chainalysis!")
            sites = sites.translate({ord('C'): None})
            
    
    df2 = df
    if "C" in sites:
        df2 = df2.iloc[: , 1:-1]
    else:
        df2 = df2.iloc[: , 1:]
    column = len(df2.columns)
    if column == 0: #jeżeli nie będzie żadnej kolumny
        print("None of the apikeys are correct!")
        raise SystemExit()      
    pm_column = []
    for c in range(column):
        for c2 in range(column):
            similar1 = similar(df2.iat[0, c], df2.iat[0, c2]) #ew. pip install thefuzz, pip install thefuzz[speedup], from thefuzz import fuzz, fuzz.ratio() 
            if (similar1 > 0.6 or similar1 == 0.0):
                pm_column.append("+")
            else:
                pm_column.append("-")
                
    result = all(element == pm_column[0] for element in pm_column)
    if (result):
       pm_column = ["+"]
    else:
       pm_column = ["-"]
                   
    df['Similarity'] = pm_column #dołączenie kolumny z podobieństwem nazw
    df['Previous Address'] = list_of_main_addresses[n] #dodanie kolumny z poprzednim adresem
    df['Epoch'] = a #dodanie kolumny z epokami    
    if a == 0:
        df.to_csv("results.txt", sep='|', mode='a', index=False, header=True) #dodanie tablicy do pliku
    else:
        df.to_csv("results.txt", sep='|', mode='a', index=False, header=False) #dodanie pustej tablicy
    
    
    if address not in temp_list_address:
        if a == (iterate - 1):
            pass
        else:
            time.sleep(5)
            hashes = [] #utworzenie listy do przechowywania hashy
            list_of_address = [] #utworzenie listy do przechowywania adresów
            loa = []
            #blockchain
            if apikey_blockcypher == "":
                hashes = hash_list_blockchain(address) #lista dla zahashowanych transakcji
                
            #blockcypher
            else:
                hashes = hash_list_blockcypher(address) #lista dla zahashowanych transakcji
        
            hashes_len = len(hashes) #pobranie dlugosci listy hashy
    
            for i in range(hashes_len): #pętla dla każdej zahashowanej transakcji w celu znalezienia adresów
                tx_output_b = [] #lista do przechowywania adresów wyciągniętych z hashy transakcji    
                time.sleep(10)
                
                #blockchain
                if apikey_blockcypher == "" or hashes_len >= 100:
                    blockchain_addresses(hashes[i], address, tx_output_b)
                    
                #blockcypher
                else:
                    blockcypher_addresses(hashes[i], address, tx_output_b)
                
                loa.extend(tx_output_b)
                unique(loa, list_of_address)
                if len(list_of_address) > max_noa: #jeżeli adres posiada posiada więcej powiązań niż z 50 innymi adresami
                    temp = [[ele for ele in sub if ele != address] for sub in temp] #usuń adres z listy adresów głównych
                    list_of_address = []
                    break
                else:
                    continue
            if list_of_address: #jeżeli lista nie jest pusta
                address_results.append(list_of_address) #dodanie listy adresów do listy adresów z danej epoki
            all_of_addresses.append(address) #dodanie do listy ze wszystkimi adresami
            finish_results.append(df) #dodanie tablicy do listy z tablicami
    else:
        pass
    temp_list_address.append(address)




address_results[0].append(main_address) #dodanie głównego adresu do listy adresów
a = -1 #ustawienie pierwszej epoki

iterate += 1  
for m in range(iterate): #pętla dla każdej epoki
    if a == -1:
        list_of_main_addresses = ["None"]
    else: #jeżeli istnieją zduplikowane adresy usuń je
        ad3 = [] #lista do przetrzymania unikatowych adresów
        temp2 = [item for sublist in temp for item in sublist]
        unique(temp2, ad3)
        list_of_main_addresses = ad3 #lista z głównymi adresami dla konkretnej epoki
            
    length = len(address_results)
    temp = []
    temp.extend(address_results) #przypisanie listy do nowej zmiennej
    address_results = [] #wyczyszczenie tablicy adresów
    temp_list_address = [] #tymczasowa tablica dla unikalnych adresów w funkcji
    a += 1 #dodanie kolejnej epoki
    for n in range(length): #pętla dla indeksu list adresów
        address_results_2 = temp[n]
        address_results_2_len = len(address_results_2)
        for o in range(address_results_2_len):
            crypto_function(address_results_2[o]) #wywołanie funkcji głównej